from rest_framework import serializers
from .transition import FinFlotransition
from .models import (
    PartyType,
    SignList,
    States,
    Flowmodel,
    TransitionManager,
    workevents,
    workflowitems,
    Action
)
from django.conf import settings
from django.apps import apps




class Workeventsserializer(serializers.ModelSerializer):
    work_model = serializers.SerializerMethodField()

    class Meta:
        model = workevents
        fields = [
            'work_model',
            'id',
            'action',
            'subaction',
            'initial_state',
            'interim_state',
            'final_state',
            'event_user',
            'record_datas',
            'from_party',
            'to_party',
            'type',
            'final_value',
            'created_date'
        ]

    def get_work_model(self,obj):
        try:
            return {"t_id" : obj.workflowitems.transitionmanager.t_id ,"model_type" : obj.workflowitems.transitionmanager.type }
        except Exception:
            return None




class Workitemserializer(serializers.ModelSerializer):
    WorkFlowEvents = Workeventsserializer(many=True, read_only=True)
    work_model = serializers.SerializerMethodField()

    class Meta:
        model = workflowitems
        fields = [
            'work_model',
            'id',
            'initial_state',
            'interim_state',
            'final_state',
            'event_user',
            'action',
            'subaction',
            'next_available_transitions',
            'model_type',
            'created_date',
            'current_from_party',
            'current_to_party',
            'WorkFlowEvents',
        ]
    

    def get_work_model(self,obj):
        try:
            return {"t_id" : obj.transitionmanager.t_id ,"model_type" : obj.transitionmanager.type }
        except Exception:
            return None




class TransitionManagerserializer(serializers.ModelSerializer):
    workflowitems = Workitemserializer(read_only = True)
    model = serializers.SerializerMethodField()
    wf_item_id = serializers.SerializerMethodField()
    class Meta:
        model = TransitionManager
        fields = [
            'id',
            'type',
            't_id',
            'in_progress',
            'wf_item_id',
            'model',
            'workflowitems'
        ]

    def get_model(self,obj):
        try:
            arr = settings.FINFLO['WORK_MODEL']
            for i in arr:
                user = apps.get_model(i)
                qs = user.objects.filter(id = obj.t_id ).values()
                if qs.exists():
                    break
            return qs
        except Exception:
            return None
                    

    def get_wf_item_id(self,obj):
        return obj.workflowitems.id



class Actionseriaizer(serializers.ModelSerializer):
    class Meta:
        model = Action
        fields = [
            'description',
            'model',
            'from_state',
            'to_state',
            'intermediator',
            'from_party',
            'to_party',
            'stage_required'
        ]



class StatesSerializer(serializers.ModelSerializer):
    class Meta:
        model = States
        fields = '__all__'


class FlowmodelsSerializer(serializers.ModelSerializer):
    class Meta:
        model = Flowmodel
        fields = '__all__'


class partytypeserializer(serializers.ModelSerializer):
    class Meta:
        model = PartyType
        fields = '__all__'



class signlistserialzier(serializers.ModelSerializer):
    class Meta:
        model = SignList
        fields = '__all__'
    

class workflowitemslistserializer(serializers.ModelSerializer):
    class Meta:
        model = workflowitems
        fields = '__all__'


class workeventslistserializer(serializers.ModelSerializer):
    class Meta:
        model = workevents
        fields = '__all__'







# Generated by Django 3.2.5 on 2022-10-17 07:40

import contextlib
from enum import unique

import django.db.models.deletion
import finflo.validators
from django.conf import settings
from django.db import migrations, models
from finflo.base_enum import Values


# RUN ON MIGRATIONS  

def create_flow_model(apps, schema_editor):
    flowmodel = apps.get_model("finflo.Flowmodel")
    try:
        a = settings.FINFLO['WORK_MODEL']
        db_alias = schema_editor.connection.alias
        for model in a:
            flowmodel.objects.using(db_alias).create(description = model.replace('.','_'))
    except:
        raise LookupError("Couldn't find work_model in settings.py check documentation")



def create_party_type_model(apps, schema_editor):
    partytype = apps.get_model("finflo.PartyType")
    with contextlib.suppress(Exception):
        abstract_models = settings.FINFLO['PARTY_MODEL']
        if abstract_models is not None:
            base_model_from_app = apps.get_model(abstract_models[0])
            for iter in base_model_from_app.objects.all():
                partytype.objects.create(description = iter.name)
    # bypassed the system for without the party_model instance


def create_sign_list(apps, schema_editor):
    states = apps.get_model("finflo.SignList")
    db_alias = schema_editor.connection.alias
    #create a default 3 signs
    states.objects.using(db_alias).bulk_create([states(name = Values.NO_SIGN.value, sign_id = 0 , sub_action_name = Values.NO_SIGN.value),
    states(name = Values.AWAITING_SIGN_A.value, sign_id = 1 , sub_action_name = Values.SIGN_A.value),
    states(name = Values.AWAITING_SIGN_B.value, sign_id = 2 , sub_action_name = Values.SIGN_B.value),
    states(name = Values.AWAITING_SIGN_C.value, sign_id = 3 , sub_action_name = Values.SIGN_C.value)])


def create_initial_draft(apps, schema_editor):
    states = apps.get_model("finflo.States")
    db_alias = schema_editor.connection.alias
    states.objects.using(db_alias).create(description = Values.DRAFT.value)
    

# DO NOT EDIT THE BELOW LINE IN PRODUCTION --
def creates_default_return_Action(apps,schema_editor):
    action = apps.get_model('finflo.Action')
    states = apps.get_model("finflo.States")
    db_alias = schema_editor.connection.alias
    action.objects.using(db_alias).create(description = Values.RETURN.value , to_state = states.objects.get(id = 1 )  )





class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='Flowmodel',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name_plural': '1. Flowmodel',
                'ordering': ['id'],
            },
        ),
        migrations.RunPython(create_flow_model),
        migrations.CreateModel(
            name='PartyType',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=255, unique=True)),
            ],
            options={
                'verbose_name_plural': '2. PartyType',
                'ordering': ['id'],
            },
        ),
        migrations.RunPython(create_party_type_model),
        migrations.CreateModel(
            name='SignList',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('sign_id', models.IntegerField(validators=[finflo.validators.validate_postive], editable = False, blank=True, null=True)),
                ('name', models.CharField(max_length=255, unique = True)),
                ('sub_action_name', models.CharField(max_length=255 , unique = True)),
            ],
            options={
                'verbose_name_plural': '5. Signatures',
                'ordering': ['id'],
            },
        ),
        migrations.RunPython(create_sign_list),
        migrations.CreateModel(
            name='States',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(max_length=255)),
            ],
            options={
                'verbose_name_plural': '6. States',
                'ordering': ['id'],
            },
        ),
        migrations.RunPython(create_initial_draft),
        migrations.CreateModel(
            name='TransitionManager',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('t_id', models.IntegerField()),
                ('type', models.CharField(max_length=255)),
                ('sub_sign', models.IntegerField(default=0, editable = False)),
                ('in_progress', models.BooleanField(default=False)),
            ],
            options={
                'verbose_name_plural': '4. TransitionManager',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='workflowitems',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('initial_state', models.CharField(default='DRAFT', max_length=50)),
                ('interim_state', models.CharField(default='DRAFT', max_length=50)),
                ('final_state', models.CharField(default='DRAFT', max_length=50)),
                ('next_available_transitions', models.JSONField(blank=True, null=True)),
                ('action', models.CharField(blank=True, default='DRAFT', max_length=25, null=True)),
                ('subaction', models.CharField(blank=True, max_length=55, null=True)),
                ('previous_action', models.CharField(blank=True, max_length=55, null=True)),
                ('model_type', models.CharField(blank=True, max_length=55, null=True)),
                ('comments', models.CharField(blank=True, max_length=500, null=True)),
                ('is_read', models.BooleanField(blank=True, default=True, null=True)),
                ('final_value', models.BooleanField(blank=True, default=False, null=True)),
                ('current_from_party', models.CharField(max_length = 500 , blank = True , null = True )),
                ('current_to_party', models.CharField(max_length = 500 , blank = True , null = True )),
                ('event_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL)),
                ('transitionmanager', models.OneToOneField(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='finflo.transitionmanager')),
            ],
            options={
                'verbose_name_plural': '7. WorkFlowItem',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='workevents',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('action', models.CharField(blank=True, default='DRAFT', max_length=25, null=True)),
                ('subaction', models.CharField(blank=True, default='DRAFT', max_length=55, null=True)),
                ('initial_state', models.CharField(default='DRAFT', max_length=50)),
                ('interim_state', models.CharField(default='DRAFT', max_length=50)),
                ('final_state', models.CharField(default='DRAFT', max_length=50)),
                ('is_read', models.BooleanField(blank=True, default=True, null=True)),
                ('record_datas', models.JSONField(blank=True, null=True)),
                ('final_value', models.BooleanField(blank=True, default=False, null=True)),
                ('comments', models.CharField(blank=True, max_length=500, null=True)),
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('type', models.CharField(blank=True, max_length=55, null=True)),
                ('event_user', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.DO_NOTHING, to=settings.AUTH_USER_MODEL)),
                ('from_party', models.CharField(max_length = 500 , blank = True , null = True )),
                ('to_party', models.CharField(max_length = 500 , blank = True , null = True )),
                ('workflowitems', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='WorkFlowEvents', to='finflo.workflowitems')),
            ],
            options={
                'verbose_name_plural': '8. WorkFlowEvent',
                'ordering': ['id'],
            },
        ),
        migrations.CreateModel(
            name='Action',
            fields=[
                ('id', models.BigAutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('description', models.CharField(blank=True, help_text='e.g., SUBMIT , DELETE', max_length=255, null=True)),
                ('intermediator', models.BooleanField(default=False)),
                ('sign_required', models.IntegerField(blank=True, default=0, editable = False, help_text='IMPORTANT : if 0 means initial_transition ', null=True)),
                ('from_party', models.ForeignKey(blank=True, help_text='this field is optional', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='from_transition_party_type', to='finflo.partytype')),
                ('from_state', models.ForeignKey(blank=True, help_text='initial from state for transition ', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='action_from_state', to='finflo.states')),
                ('model', models.ForeignKey(blank=True, null=True, on_delete=django.db.models.deletion.CASCADE, to='finflo.flowmodel')),
                ('stage_required', models.ForeignKey(blank=True, help_text='this field is optional , IMPORTANT : if NO_SIGN means initial_transition', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='finflo.signlist')),
                ('to_party', models.ForeignKey(blank=True, help_text='this field is optional', null=True, on_delete=django.db.models.deletion.DO_NOTHING, related_name='to_transition_party_type', to='finflo.partytype')),
                ('to_state', models.ForeignKey(blank=True, help_text='final state for the transition to take place', null=True, on_delete=django.db.models.deletion.DO_NOTHING, to='finflo.states')),
            ],
            options={
                'verbose_name_plural': '3. Action',
                'ordering': ['id'],
                'unique_together': {('description', 'model')},
            },
        ),
        migrations.RunPython(creates_default_return_Action),
    ]



from django.contrib import admin
from .models import Action, Flowmodel, PartyType, States, TransitionManager , workevents , workflowitems , SignList 
# Register your models here.
from django.contrib import messages
from django.utils.translation import ngettext
from django.conf import settings

@admin.action(description='Reset selected items')
def make_in_progress_to_false(modeladmin, request, queryset):
    queryset.update(in_progress = False , sub_sign = 0)


class TransitionModeladmin(admin.ModelAdmin):
    list_display = ("t_id", "type", "in_progress")
    actions = [make_in_progress_to_false]


class CustomActionAdminModel(admin.ModelAdmin):

    def has_delete_permission(self, request, obj=None):
        return obj is None or obj.pk != 1

class customadminforsign(admin.ModelAdmin):

    def has_delete_permission(self, request, obj=None):
        return False
    

admin.site.register(Flowmodel)
admin.site.register(TransitionManager,TransitionModeladmin)
admin.site.register(Action , CustomActionAdminModel )
admin.site.register(States)
admin.site.register(PartyType)
admin.site.register(SignList , customadminforsign)
admin.site.register(workflowitems)
admin.site.register(workevents)


import contextlib
from django import dispatch
from django.db.models.signals import post_save  , post_delete
from .exception import DefaultsNotFoundError
from .middleware import get_current_user
from .models import (
    PartyType,
    States,
    TransitionManager,
    workevents,
    workflowitems 
)
from django.conf import settings
from .transition import FinFlotransition
from .compatability import check_version_compatibility



# MAIN SIGNAL RECEIVER FOR TRANSITION HANDLING - 4/8/2022 by anand

Base_link = 'https://pypi.org/project/finflo/'


# CORE LIST ITERATED FROM THE DEFAULT SETTINGS -- PRODUCTION : 27/10/2022
try:
    check_version_compatibility()
    WORK_MODEL_SENDER = list(settings.FINFLO['WORK_MODEL'])
    PARTY_MODEL_SENDER = list(settings.FINFLO['PARTY_MODEL'])
except:
    raise DefaultsNotFoundError ("unable to find WORK_MODEL and PARTY_MODEL in your settings.py file , \
        check the documentation for more details",Base_link)




# CUSTOM MULTIPLE RECEIVER SIGNALS 

def finflo_receiver(signal, senders ,  **kwargs):
    def decorator(receiver_func):
        for sender in senders:
            if isinstance(signal, (list, tuple)):
                for s in signal:
                    s.connect(receiver_func, sender=sender, **kwargs)
            else:
                signal.connect(receiver_func, sender=sender,**kwargs)
        return receiver_func
    return decorator




# MULTIPLE RECEIVER MODEL WITH MULTIPLE SENDERS

#### CREATER

## A
@finflo_receiver(post_save, senders = WORK_MODEL_SENDER)
def creates_finflo_work_models(sender, instance, created , **kwargs):
    if created:
        model = sender._meta.db_table
        states = States.objects.get(id = 1)
        obj = TransitionManager.objects.create(type = model , t_id = instance.id)
        gets_transition = FinFlotransition(t_id = obj.id , type =   model )
        obj2 = workflowitems.objects.create(transitionmanager = obj ,model_type = model , initial_state = states.description ,interim_state = states.description, final_state = states.description)
        workevents.objects.create(workflowitems = obj2 , type = model , record_datas = gets_transition.get_record_datas(), initial_state = states.description ,interim_state = states.description,  final_state = states.description)
        

## B
@finflo_receiver(post_save, senders = PARTY_MODEL_SENDER)
def creates_party_type(sender, instance, **kwargs):
    with contextlib.suppress(Exception):
        PartyType.objects.create(description = instance.name)



# POST DELETE 

## A
@finflo_receiver(post_delete, senders = WORK_MODEL_SENDER)
def delete(sender, instance, **kwargs):
    TransitionManager.objects.filter(type = sender._meta.db_table , t_id = instance.id).delete()
    

##  B
@finflo_receiver(post_delete, senders = PARTY_MODEL_SENDER)
def delete(sender, instance, **kwargs):
    PartyType.objects.delete(description = instance.name)





import json
from .models import (
    SignList,
    TransitionManager,
    Action,
    workevents,
    workflowitems,
    Flowmodel
)
from django.core import serializers as core_serializers
from .middleware import get_current_user
from .exception import (
    ModelNotFound,
    TransitionNotAllowed,
    ReturnModelNotFound,
)
from django.db.models import Q
from django.conf import settings
from collections import deque
from django.apps import apps
from django.forms.models import model_to_dict




####################################################
#############       CORE     #######################
####################################################


class FinFlotransition():

    # BASE CONSTRUCTORS # 
    
    def __init__(self, type, t_id , action = None ,source = None,interim = None, target = None, from_party = None, to_party = None):
        # typing support needed 
        self.type = type
        self.t_id = t_id
        self.action = action.upper() if action else None
        self.source = source or None
        self.interim = interim or None
        self.target = target or None
        self.from_party = from_party or None
        self.to_party = to_party or None
        gets_current_Action = None
        gets_return_action = self.gets_default_return_action()
        if action:
            if action == gets_return_action.description :
                self.return_transition()
            else:
                self.transition()
        else:
            self.manualtransitions()
        # return None

    def __repr__(self):
        return f"the id is {self.t_id} and type is {self.type}"

    
    def __str__(self):
        return f"the id is {self.t_id} and type is {self.type}"
    


    


    # GETS THE ALL MODEL VALUE FOR TRANSITION #

    def gets_base_action(self):
        try:
            return Action.objects.get(description=self.action, model__description = self.type )
        except:
            raise ModelNotFound()
    
        
    # RETURN ACTION
    def gets_default_return_action(self):
        try:
            return Action.objects.get(id = 1)
        except:
            raise ReturnModelNotFound()

        

    # GETS WORKFLOW MODEL ID #
    def gets_wf_item(self , model):
        ws = workflowitems.objects.get(transitionmanager=model)
        return ws
    

    # GETS TRANSITION MANAGER

    def gets_base_model(self):
        try:
            return TransitionManager.objects.get(type = self.type , t_id = self.t_id)
        except Exception:
            ModelNotFound()

    
    # GETS ALL ACTION
    def gets_all_models(self):
        try:
            gets_model = TransitionManager.objects.get(type = self.type , t_id = self.t_id)
            gets_flows = Flowmodel.objects.get(description  = self.type)
            gets_action = Action.objects.get(Q(model = gets_flows.id ) | Q(model = None), description=self.action)
            gets_wf  = self.gets_wf_item(gets_model.id)
            sign_lists = sub_action_list = []
            try:
                for item in SignList.objects.all():
                    sign_lists.append(item.name)
                    if item.name == gets_action.stage_required.name :
                        break
                next_avail_trans = sign_lists[gets_model.sub_sign : ]
                next_avail_trans_value = deque(next_avail_trans)
                next_avail_trans_value.popleft()
                next_states = list(next_avail_trans_value)
                next_avail_Transition = {'values' : next_states}
            except Exception:
                next_avail_Transition = None
            return gets_model , gets_action , gets_flows , gets_wf , sign_lists , next_avail_Transition
        except:
            raise ModelNotFound()

        

    def return_transition(self):
        
        overall_model = self.gets_all_models()
        obj, created = workflowitems.objects.update_or_create(
            transitionmanager=overall_model[0] or overall_model[0].id,
            defaults={
                "initial_state": overall_model[1].from_state.description
                if overall_model[1].from_state
                else overall_model[3].initial_state,
                "interim_state": overall_model[1].to_state.description,
                "final_state": overall_model[1].to_state.description,
                "action": self.action,
                "subaction": self.action,
                'previous_action': overall_model[3].action,
                "next_available_transitions": None,
                "model_type": self.type,
                "event_user": get_current_user(),
                "final_value": True,
                "current_from_party": overall_model[1].from_party
                or self.from_party,
                "current_to_party": overall_model[1].to_party or self.to_party,
            },
        )
        workevents.objects.create(
            workflowitems=overall_model[3],
            event_user=get_current_user(),
            initial_state=overall_model[1].from_state.description
            if overall_model[1].from_state
            else overall_model[3].initial_state,
            final_value=True,
            record_datas=self.get_record_datas(),
            interim_state=overall_model[1].to_state.description,
            final_state=overall_model[1].to_state.description,
            action=self.action,
            subaction=self.action,
            type=self.type,
            from_party=overall_model[1].from_party or self.from_party,
            to_party=overall_model[1].to_party or self.to_party,
        )
        overall_model[0].sub_sign = 0
        overall_model[0].in_progress = False
        overall_model[0].save()                      

    

    def get_record_datas(self):
        overall_model = self.gets_base_model()
        try:
            work_model = settings.FINFLO['WORK_MODEL']  
            for iter in work_model:
                gets_model =  apps.get_model(iter)
                query_data = gets_model.objects.filter(id = overall_model.t_id)
                base_serialized_Data = core_serializers.serialize('json', query_data)
                if query_data.exists() and (gets_model._meta.db_table == overall_model.type):
                    break
                continue
            return {"values" : json.loads(base_serialized_Data)}
        except:
            return {"values" : None}
           

   

    # MANUAL TRANSITION WITH SOURCE , INTERIM , AND TARGET STATES

    def manualtransitions(self):
        try:
            queryset = self.gets_base_model()
            obj , created  = workflowitems.objects.update_or_create( transitionmanager= queryset , defaults= {"initial_state" : self.source, "interim_state" : self.interim ,
                    "final_state" :  self.target  , "next_available_transitions" : None, "model_type" : self.type, "event_user" : get_current_user() , "current_from_party" : self.from_party, "current_to_party" : self.to_party ,"final_value" : True})
            workevents.objects.create(workflowitems = obj , event_user=get_current_user(),  initial_state=self.source  , 
                    interim_state = self.interim , final_state=self.target ,type=self.type , record_datas = self.get_record_datas() , from_party = self.from_party , to_party = self.to_party , final_value = True)
        except:
            return None  
       



    ## CORE TRANSITION ###

    def transition(self):
       
        overall_model = self.gets_all_models()

        if overall_model[0] is not None:
            def Transition_Handler():
                
                if len(overall_model[4]) == overall_model[0].sub_sign:
                    return
                try:
                    ws = workflowitems.objects.update_or_create(
                        transitionmanager=overall_model[0] or overall_model[0].id,
                        defaults={
                            "initial_state": overall_model[
                                1
                            ].from_state.description,
                            "interim_state": overall_model[4][
                                1 + overall_model[0].sub_sign
                            ],
                            "final_state": overall_model[1].to_state.description,
                            "next_available_transitions": overall_model[5],
                            "action": self.action,
                            "subaction": overall_model[4][
                                overall_model[0].sub_sign
                            ],
                            "model_type": self.type,
                            "event_user": get_current_user(),
                            "current_from_party": overall_model[1].from_party
                            or self.from_party,
                            "current_to_party": overall_model[1].to_party
                            or self.to_party,
                        },
                    )
                    workevents.objects.create(
                        workflowitems=overall_model[3],
                        event_user=get_current_user(),
                        initial_state=overall_model[1].from_state.description,
                        interim_state=overall_model[4][
                            1 + overall_model[0].sub_sign
                        ],
                        record_datas=self.get_record_datas(),
                        final_state=overall_model[1].to_state.description,
                        action=self.action,
                        subaction=self.action,
                        type=self.type,
                        from_party=overall_model[1].from_party or self.from_party,
                        to_party=overall_model[1].to_party or self.to_party,
                    )
                    overall_model[0].sub_sign += 1
                    overall_model[0].in_progress = True
                    overall_model[0].save()
                except Exception:
                    ws = workflowitems.objects.update_or_create(
                        transitionmanager=overall_model[0] or overall_model[0].id,
                        defaults={
                            "initial_state": overall_model[
                                1
                            ].from_state.description,
                            "interim_state": overall_model[1].to_state.description,
                            "final_state": overall_model[1].to_state.description,
                            "action": self.action,
                            "next_available_transitions": None,
                            "subaction": self.action,
                            "model_type": self.type,
                            "event_user": get_current_user(),
                            "current_from_party": overall_model[1].from_party
                            or self.from_party,
                            "current_to_party": overall_model[1].to_party
                            or self.to_party,
                            "final_value": True,
                        },
                    )
                    workevents.objects.create(
                        workflowitems=overall_model[3],
                        event_user=get_current_user(),
                        initial_state=overall_model[1].from_state.description,
                        interim_state=overall_model[1].to_state.description,
                        record_datas=self.get_record_datas(),
                        final_state=overall_model[1].to_state.description,
                        action=self.action,
                        subaction=self.action,
                        type=self.type,
                        from_party=overall_model[1].from_party or self.from_party,
                        to_party=overall_model[1].to_party or self.to_party,
                        final_value=True,
                    )
                    overall_model[0].sub_sign = 0
                    overall_model[0].in_progress = False
                    overall_model[0].save()

            return Transition_Handler()
        else:
            raise TransitionNotAllowed()



class ReverseTransition:
    pass


# Generated by Django 3.2.5 on 2022-10-14 10:50

import contextlib
from django.conf import settings
from django.db import migrations




# RUN ON MIGRATIONS  -- remigrate if work_model is updated 

def update_flow_model(apps, schema_editor):
    flowmodel = apps.get_model("finflo.Flowmodel")
    try:
        a = settings.FINFLO['WORK_MODEL']
        db_alias = schema_editor.connection.alias
        for model in a:
            print("FLOWS MIGRATING")
            flowmodel.objects.using(db_alias).update_or_create(description = model.replace('.','_'))
    except:
        raise LookupError("Couldn't find work_model in settings.py check documentation")


def updates_party_type_model(apps, schema_editor):
    partytype = apps.get_model("finflo.PartyType")
    abstract_models = settings.FINFLO['PARTY_MODEL']
    db_alias = schema_editor.connection.alias
    with contextlib.suppress(Exception):
        if abstract_models:
            base_model_from_app = apps.get_model(abstract_models[0])
            for iter in base_model_from_app.objects.all():
                partytype.objects.using(db_alias).update_or_create(description = iter.name)
            print("PARTY MODELS CREATED SUCCESSFULLY")


    
class Migration(migrations.Migration):

    dependencies = [
        ('finflo', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(update_flow_model),
        migrations.RunPython(updates_party_type_model)
    ]


from rest_framework.views import APIView
from rest_framework import status
from rest_framework.response import Response
from rest_framework.permissions import IsAuthenticated
from .transition import FinFlotransition
from rest_framework.generics import (
    ListAPIView,
    ListCreateAPIView,
    RetrieveUpdateAPIView
)
from django.shortcuts import get_object_or_404
from django.conf import settings
from .serializer import  (
    StatesSerializer,
    TransitionManagerserializer,
    Actionseriaizer,
    partytypeserializer,
    signlistserialzier,
    FlowmodelsSerializer,
    workeventslistserializer,
    Workitemserializer
)
from .models import (
    Action, 
    PartyType,
    States,
    SignList,
    Flowmodel,
    TransitionManager, 
    workevents, 
    workflowitems
)
from django.db.models import Q



####################################################
################      API       ####################
####################################################



# 1 . TRANSITION MAKING API


class TransitionApiView(APIView):
    permission_classes = [IsAuthenticated]
    
    def post(self,request):
        type = request.data.get("type")
        action = request.data.get("action")
        t_id = request.data.get("t_id")
        # extra datas for manual transitions
        source = request.data.get("source")
        interim = request.data.get("interim")
        target = request.data.get("target")
        from_party = request.data.get("from_party")
        to_party = request.data.get("to_party")

        if type and t_id  is not None:
            transitions = FinFlotransition(action = action , type = type , t_id = t_id , source = source ,
            interim = interim ,target = target ,from_party = from_party , to_party = to_party )
            return Response({"status" : "Transition success"},status = status.HTTP_200_OK)
        return Response({"status" : "Transition failure"},status = status.HTTP_204_NO_CONTENT)


class TransitionResetApiview(APIView):
    permission_classes = [IsAuthenticated]

    def post(self,request):
        t_id = request.data.get("t_id")
        type = request.data.get("type")
        if type and t_id : 
            TransitionManager.objects.filter(t_id = t_id , type = type).update(in_progress = False)
            return Response({"status" : "Status updated successfully"},status = status.HTTP_200_OK)
        return Response({"status" : "Failure"},status = status.HTTP_204_NO_CONTENT)

        



#  2 . ALL WORK_MODEL LIST 


class DetailsListApiView(ListAPIView):
    queryset = TransitionManager.objects.all()
    serializer_class = TransitionManagerserializer
    permission_classes = [IsAuthenticated]
    
    def get_queryset(self):
        # sourcery skip: lift-return-into-if, remove-redundant-if
        type = self.request.query_params.get('type',None)
        t_id = self.request.query_params.get('t_id',None)
        if t_id :
            queryset = TransitionManager.objects.filter(t_id = t_id)
        elif type and t_id :
            queryset = TransitionManager.objects.filter(type__icontains = type , t_id = t_id)
        else:
            queryset = TransitionManager.objects.all()
        return queryset

    def list(self, request):
        queryset = self.get_queryset()
        serializer = TransitionManagerserializer(queryset, many=True)
        return Response({"status": "success", "type" : settings.FINFLO['WORK_MODEL'] , "data": serializer.data}, status=status.HTTP_200_OK)



# 3 . WORFLOW API 

class WorkFlowitemsListApi(RetrieveUpdateAPIView):
    queryset = workflowitems.objects.all()
    serializer_class = Workitemserializer
    permission_classes = [IsAuthenticated]


    def retrieve(self, request, pk=None):
        queryset = workflowitems.objects.all()
        user = get_object_or_404(queryset, pk=pk)
        serializer = Workitemserializer(user)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)



# WORKEVENTS API 


class WorkEventsListApi(RetrieveUpdateAPIView):
    queryset = workevents.objects.all()
    serializer_class = workeventslistserializer
    permission_classes = [IsAuthenticated]

    def retrieve(self, request, pk=None):
        queryset = workevents.objects.all()
        user = get_object_or_404(queryset, pk=pk)
        serializer = workeventslistserializer(user)
        return Response({"status": "success", "data": serializer.data}, status=status.HTTP_200_OK)



# ACTION CREATE AND LIST API 


class ActionListApi(ListCreateAPIView):
    queryset = Action.objects.all()
    serializer_class = Actionseriaizer
    permission_classes = [IsAuthenticated]

    def get_queryset(self):
        action = self.request.query_params.get('action',None)
        model = self.request.query_params.get('model',None)
        if (action and model) is not None:
            return Action.objects.filter(
                description__icontains=action, model__description__icontains=model
            )
        elif action:
            return Action.objects.filter(description__icontains = action)
        else:
            return Action.objects.all()

    def list(self, request):
        queryset = self.get_queryset()
        serializer = Actionseriaizer(queryset, many=True)
        return Response({"status": "success" , "data": serializer.data}, status=status.HTTP_200_OK)


    def post(self, request):
        serializer = Actionseriaizer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response({"status": "success", "data": serializer.data}, status=status.HTTP_201_CREATED)
        return Response({"status": "failure", "data": serializer.errors},status=status.HTTP_204_NO_CONTENT)


# STATES API 


class statesListCreateApi(ListCreateAPIView):
    queryset = States.objects.all()
    serializer_class = StatesSerializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = States.objects.all()
        serializer = StatesSerializer(queryset, many=True)
        return Response({"status": "success" , "data": serializer.data}, status=status.HTTP_200_OK)


    def post(self, request):
        serializer = StatesSerializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response({"status": "success", "data": serializer.data}, status=status.HTTP_201_CREATED)
        return Response({"status": "failure", "data": serializer.errors},status=status.HTTP_204_NO_CONTENT)



# FLOW MODEL GET API 

class FlowModelGetApi(ListCreateAPIView):
    queryset = Flowmodel.objects.all()
    serializer_class = FlowmodelsSerializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = Flowmodel.objects.all()
        serializer = FlowmodelsSerializer(queryset, many=True)
        return Response({"status": "success" , "data": serializer.data}, status=status.HTTP_200_OK)



# PARTY TYPE API 

class PartyTypeListCreateApi(ListCreateAPIView):
    queryset = PartyType.objects.all()
    serializer_class = partytypeserializer
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = PartyType.objects.all()
        serializer = partytypeserializer(queryset, many=True)
        return Response({"status": "success" , "data": serializer.data}, status=status.HTTP_200_OK)


    def post(self, request):
        serializer = partytypeserializer(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response({"status": "success", "data": serializer.data}, status=status.HTTP_201_CREATED)
        return Response({"status": "failure", "data": serializer.errors},status=status.HTTP_204_NO_CONTENT)



# SIGN LIST API 


class SignListListCreateApi(ListCreateAPIView):
    queryset = SignList.objects.all()
    serializer_class = signlistserialzier
    permission_classes = [IsAuthenticated]

    def list(self, request):
        queryset = SignList.objects.all()
        serializer = signlistserialzier(queryset, many=True)
        return Response({"status": "success" , "data": serializer.data}, status=status.HTTP_200_OK)

    def post(self, request):
        serializer = signlistserialzier(data=request.data)
        if serializer.is_valid():
            serializer.save()
            return Response({"status": "success", "data": serializer.data}, status=status.HTTP_201_CREATED)
        return Response({"status": "failure", "data": serializer.errors},status=status.HTTP_204_NO_CONTENT)